/* -------------------------------------------------------------------------------------------------------------------------
* Project           :C-to-FPGA Solution for Acclerating Tomographic Reconstruction
* File Name         :uwer_SystemMatrix.c (from project UW Emission Reconstruction) 
*                    Copyright (c) 2004 University of Washington - Imaging Research Laboratory. All rights reserved.
* Version           :1.0 (V1)
* (V1)Author        :Adam Alessio
* (V1)Created       :Wed, Jan 14 2004
* References        :
* Description       :
*
******This file may not be distributed or modified without this header*****
*-------------------------------------------------------------------------------------------------------------------------*/


#include "uwer_SystemMatrix.h"
#define PRECOMPUTE_P_MATRIX
#define PIX_SUBSAMPLE 15;
#define LEN 512;

/************************************************************************************************/
/* Function: Orchestrates the generation of the system matrix P */
//
// Will either compute and store P once, or compute P on the fly
//
// y = P * x
// where y is M x 1 vector of the data,
// P is M x N system matrix and
// x is N x 1 vector of the image.
//
// P_ij is the probability that an emission from voxel j
// is detected in LOR i.
/**/

void generate_system_matrix(struct sino_info *sinoInfo,
			    struct recon_image_info *imageInfo,
			    struct system_matrix_info *systemInfo,
			    struct exec_parameters *params)
{
  double xPixelSize,yPixelSize,tfov, val;
  double **pixelCont;
  int j1,j2,r;
  int subSet,numSubSets,*nRowsPerSubset, cnts;
  struct p_column_dbl one_p_column;
  float *attenTemp;
  FILE *fp;
  
#ifdef PRECOMPUTE_P_MATRIX      // Compute and store the system matrix once
  xPixelSize= imageInfo->xPixelSize;
  yPixelSize= imageInfo->yPixelSize;
  tfov= imageInfo->tfov;
  
  Print("Generating System Matrix\n");
  
  find_pixel_contributions(sinoInfo, imageInfo,&pixelCont);  // fill pixelCont
  
  numSubSets = imageInfo->nSubSets;
  one_p_column.val = (double *) get_spc(sinoInfo->nProjs*sinoInfo->nAngles, sizeof(double));
  one_p_column.row = (int *) get_spc(sinoInfo->nProjs*sinoInfo->nAngles, sizeof(int));
  nRowsPerSubset = (int *) get_spc(numSubSets, sizeof(int));

  // Load Attenuation Correction Information

  if(!params->attenCorrect_flag)
    //  Print("Not applying attenuation correction in system matrix\n");
    Print(" ");
  else
    {
      
      Print("\n------------\nNotice:::  Applying Attenuation Correction in system matrix\n");
      if((fp = fopen(params->AttenCorrectFileName,"r"))==NULL){
	Print("\nCan not read attenuation input file %sn\n\a\n", params->AttenCorrectFileName);
	exit(-3);
      }
      
      attenTemp = (float*)get_spc(sinoInfo->nAngles*sinoInfo->nProjs,sizeof(float)); 
      cnts = fread(attenTemp,sizeof(float), sinoInfo->nAngles*sinoInfo->nProjs,fp);
      fclose(fp);
    }

  
  // Allocate initial space for system matrix
  // systemInfo->p_matrix = (struct p_column_flt ***)get_spc(numSubSets, sizeof(struct p_column_flt **));
  // for(subSet = 0;subSet<numSubSets;subSet++)
  // systemInfo->p_matrix[subSet] = (struct p_column_flt **)multialloc(sizeof(struct p_column_flt),2,imageInfo->yDim,imageInfo->xDim);
    
  systemInfo->p_matrix = (struct p_column_flt ***)multialloc(sizeof(struct p_column_flt),3,numSubSets,imageInfo->yDim,imageInfo->xDim);
      
  for (j1=0;j1<imageInfo->yDim;j1++)
    for (j2=0;j2<imageInfo->xDim;j2++)
      {
	// printf("column for %d, %d\n",j1,j2);
	get_one_p_column_for_j(j1,j2,xPixelSize, yPixelSize,tfov,sinoInfo,imageInfo,&pixelCont,&one_p_column);
	//printf("finished this column\n");

	// Re-Initialize counter
	for(subSet = 0;subSet<numSubSets;subSet++)
	  nRowsPerSubset[subSet]=0;
	
	// Find number of non-zero projections in this subset	    
	for (r=0;r<one_p_column.nRows;r++){
	  subSet = one_p_column.row[r]%numSubSets;
	  nRowsPerSubset[subSet]++;
	}
	// Allocate space for each subset column
	for(subSet = 0;subSet<numSubSets;subSet++){
	  systemInfo->p_matrix[subSet][j1][j2].val=(float *) get_spc(nRowsPerSubset[subSet], sizeof(float));
	      systemInfo->p_matrix[subSet][j1][j2].row = (unsigned int *) get_spc(nRowsPerSubset[subSet], sizeof(unsigned int));
	      systemInfo->p_matrix[subSet][j1][j2].nRows = nRowsPerSubset[subSet];
	      nRowsPerSubset[subSet] = 0;  // Initialize to 0 to index in next loop
	      systemInfo->p_matrix[subSet][j1][j2].norm = 0.0;
	    }
	// Place projections in the correct subset
	// Do not do AW
	if(!params->attenCorrect_flag){
	  for (r=0;r<one_p_column.nRows;r++){
	    subSet = one_p_column.row[r]%numSubSets;
	    systemInfo->p_matrix[subSet][j1][j2].val[nRowsPerSubset[subSet]] = (float)one_p_column.val[r];
	    systemInfo->p_matrix[subSet][j1][j2].norm += one_p_column.val[r];
	    systemInfo->p_matrix[subSet][j1][j2].row[nRowsPerSubset[subSet]++] = (unsigned int)one_p_column.row[r];
	  }
	}
	else   // Do AW
	  {
	    for (r=0;r<one_p_column.nRows;r++){
	      subSet = one_p_column.row[r]%numSubSets;
	      val =  one_p_column.val[r]*(double)attenTemp[r];
	      systemInfo->p_matrix[subSet][j1][j2].val[nRowsPerSubset[subSet]] = (float)val;
	      systemInfo->p_matrix[subSet][j1][j2].norm += val;
	      systemInfo->p_matrix[subSet][j1][j2].row[nRowsPerSubset[subSet]++] = (unsigned int)one_p_column.row[r];
	    }
	  }
	
	    // For no subsets
	   
		// for (r=0;r<one_p_column.nRows;r++)
		// {
		// systemInfo->p_matrix[j1][j2].val[r] = (float)one_p_column.val[r];   
		// systemInfo->p_matrix[j1][j2].row[r] = (unsigned int)one_p_column.row[r];
		// }
	  }


  /* TEST SYSTEM MATRIX */
  // for (j1=0;j1<imageInfo->yDim;j1++)
  // for (j2=0;j2<imageInfo->xDim;j2++)
  //  for(subSet = 0;subSet<numSubSets;subSet++){
  //if(systemInfo->p_matrix[subSet][j1][j2].nRows)
  //  if(systemInfo->p_matrix[subSet][j1][j2].norm == 0.0)
  //    printf("j1 = %d, j2 = %d, subSet = %d, norm = %f, nRows = %d\n", j1,j2,subSet,systemInfo->p_matrix[subSet][j1][j2].norm,systemInfo->p_matrix[subSet][j1][j2].nRows);
  //  }
  
	
      free((void*)nRowsPerSubset);
      free((void*)one_p_column.val);
      free((void*)one_p_column.row);
      if(params->attenCorrect_flag)
	free((void*)attenTemp);

      
      fflush(stderr);
      Print("Finished generating System Matrix\n");
      fflush(stdout);
#endif
      
}  // End of generate_system_matrix()

/************************************************************************************************/
/* Function: Orchestrates the generation of the system matrix P */
//
//  Designed for 2+1 PSF (variant in s and influential in s and z)

void generate_system_matrix_PSF21(struct sino_info *sinoInfo,
		     struct recon_image_info *imageInfo,
				  struct system_matrix_info *systemInfo,struct exec_parameters *params)
{
  double xPixelSize,yPixelSize,tfov;
  double **pixelCont;
  int jj1,jj2,r, zDim,row, zsrf, ssrf, completeS, sDim, nProjs,sStart;
  int subset,numSubSets,**nRows, sinoSize, cnts;
  float **kernel;
  double **blurredSpace, val; 
  struct p_column_dbl one_p_column;
  int **testCount;
  float *attenTemp;
  FILE *fp;
  
  systemInfo->systemResponse_flag = 0;
  
  xPixelSize= imageInfo->xPixelSize;
  yPixelSize= imageInfo->yPixelSize;
  tfov= imageInfo->tfov;
  
  Print("Generating System Matrix\n");
  
  find_pixel_contributions(sinoInfo, imageInfo,&pixelCont);  // fill pixelCont

  numSubSets = imageInfo->nSubSets;
  one_p_column.val = (double *) get_spc((int)(sinoInfo->nProjs*sinoInfo->nAngles), sizeof(double));
  one_p_column.row = (int *) get_spc((int)(sinoInfo->nProjs*sinoInfo->nAngles), sizeof(int));


  if(!params->attenCorrect_flag)
    // Print("Not applying attenuation correction in system matrix\n");
    Print(" ");
  else
    {
      
      Print("\n------------\nNotice:::  Applying Attenuation Correction in system matrix\n");
      if((fp = fopen(params->AttenCorrectFileName,"r"))==NULL){
	Print("\nCan not read attenuation input file %sn\n\a\n", params->AttenCorrectFileName);
	exit(-3);
      }
      
      attenTemp = (float*)get_spc(sinoInfo->nAngles*sinoInfo->nProjs,sizeof(float)); 
      cnts = fread(attenTemp,sizeof(float), sinoInfo->nAngles*sinoInfo->nProjs,fp);
      fclose(fp);
    }
  
  // Allocate initial space for system matrix
  // systemInfo->p_matrix = (struct p_column_flt ***)get_spc(numSubSets, sizeof(struct p_column_flt **));
  // for(subSet = 0;subSet<numSubSets;subSet++)
  // systemInfo->p_matrix[subSet] = (struct p_column_flt **)multialloc(sizeof(struct p_column_flt),2,imageInfo->yDim,imageInfo->xDim);
    
  sDim = systemInfo->srfDim[2];
  zDim = systemInfo->srfDim[1];
  nProjs = sinoInfo->nProjs;
  sinoSize = nProjs*sinoInfo->nAngles;
  
  nRows = (int**)multialloc(sizeof(int),2,zDim, numSubSets);
  testCount = (int**)multialloc(sizeof(int),2,zDim, numSubSets);

  sStart = -(int)(systemInfo->srfDim[2]/2);
  
  blurredSpace = (double**)multialloc(sizeof(double),2,zDim,sinoSize);
  for(zsrf= 0;zsrf<zDim;zsrf++)
    for(r=0;r<sinoSize;r++)
      blurredSpace[zsrf][r] = 0.0;

  
  systemInfo->p_matrix = (struct p_column_flt***)multialloc(sizeof(struct p_column_flt),3,(int)(imageInfo->yDim),(int)(imageInfo->xDim),(int)(numSubSets*zDim));
  
  for (jj1=0;jj1<imageInfo->yDim;jj1++){
    if(jj1 % 10 == 0)
      printf("Pixel Row:%d\n", jj1);
    
    for (jj2=0;jj2<imageInfo->xDim;jj2++)
      {

	//	break;  /** AMA TEST */
	// initialize nRows
	for(zsrf = 0;zsrf<zDim;zsrf++)
	  for(subset=0;subset<numSubSets;subset++)
	    nRows[zsrf][subset] = 0;

	// Get Unblurred System column
	get_one_p_column_for_j(jj1,jj2,xPixelSize, yPixelSize,tfov,sinoInfo,imageInfo,&pixelCont,&one_p_column);


	// Do not do AW
	if(!params->attenCorrect_flag){
	// Blur with PSF
	for (r=0;r<one_p_column.nRows;r++){
	  val = one_p_column.val[r];
	  row = one_p_column.row[r];
	  kernel = (float**)systemInfo->srf[row%nProjs];   // Choose correct kernel which is variant in s
	  for(ssrf = 0;ssrf<sDim;ssrf++){
	    completeS = row+sStart+ssrf;      // This will still have wrap-around in sinogram, but not going to worry about now
	    if(completeS>=0 && completeS < sinoSize){
	      for(zsrf = 0;zsrf<zDim;zsrf++){   // Blur this projection in z and s
	/* 	printf("row = %d, zsrf = %d, ssrf = %d, kernel = %f\n", row, zsrf, ssrf,(double)kernel[zsrf][ssrf]); */
		blurredSpace[zsrf][completeS] += (double)kernel[zsrf][ssrf]*val;
	      }}}}
	}
	else // Do AW
	  {
	    for (r=0;r<one_p_column.nRows;r++){
	      val = one_p_column.val[r]*(double)attenTemp[r];
	      row = one_p_column.row[r];
	      kernel = (float**)systemInfo->srf[row%nProjs];   // Choose correct kernel which is variant in s
	      for(ssrf = 0;ssrf<sDim;ssrf++){
		completeS = row+sStart+ssrf;      // This will still have wrap-around in sinogram, but not going to worry about now
		if(completeS>=0 && completeS < sinoSize){
		  for(zsrf = 0;zsrf<zDim;zsrf++){   // Blur this projection in z and s
		    blurredSpace[zsrf][completeS] += (double)kernel[zsrf][ssrf]*val;
		  }}}}
	  }

	// Find Size in each subset:

	for(zsrf = 0;zsrf<zDim;zsrf++)
	  for(r=0;r<sinoSize;r++){
	    if(blurredSpace[zsrf][r]!=0.0){
	      (nRows[zsrf][r%numSubSets])+=1;
	    }}  
      
	// Allocate space for each subset column 
	for(subset=0;subset<numSubSets;subset++)
	  for(zsrf = 0;zsrf<zDim;zsrf++){	    
	    completeS = subset*zDim+zsrf;
	    systemInfo->p_matrix[jj1][jj2][completeS].nRows = nRows[zsrf][subset];
	    systemInfo->p_matrix[jj1][jj2][completeS].norm = 0.0;
	    if(nRows[zsrf][subset]!=0)
	      {
		systemInfo->p_matrix[jj1][jj2][completeS].val=(float*)get_spc((int)(nRows[zsrf][subset]), sizeof(float));
		systemInfo->p_matrix[jj1][jj2][completeS].row = (int*)get_spc((int)(nRows[zsrf][subset]), sizeof(int));
	      }
	     
	    /*	ind = 0;
	// Place projections in the sparse form

	for(r=subset;r<sinoSize;r+=numSubSets){  // Requires numSubSets to be a factor of nProjs
	  val =blurredSpace[zsrf][r];

	  if(val!=0.0){ // Not equivalency test, just assignment
	    systemInfo->p_matrix[completeS][jj1][jj2].val[ind] = (float)val;
	    systemInfo->p_matrix[completeS][jj1][jj2].row[ind++] = r;
	    systemInfo->p_matrix[completeS][jj1][jj2].norm +=(double)val;
	    blurredSpace[zsrf][r] = 0.0;
	    }
if(ind!=nRows[zsrf][subset])
	  printf("Sparse Error at zsrf = %d, subset = %d, ind = %d, nRows = %d\n", zsrf, subset, ind, nRows[zsrf][subset]);
	    */
	  }
	    
	// Brute force to find bug
	for(zsrf = 0;zsrf<zDim;zsrf++)
	  for(subset=0;subset<numSubSets;subset++)
	    testCount[zsrf][subset] = 0;

	
	for(zsrf = 0;zsrf<zDim;zsrf++)
	  for(r=0;r<sinoSize;r++){
	    val = blurredSpace[zsrf][r];
	    subset =(int)(r%numSubSets);
	    if(val!=0.0){
	      completeS = (subset)*zDim+zsrf;
	      systemInfo->p_matrix[jj1][jj2][completeS].val[testCount[zsrf][subset]] = (float)val;
	      systemInfo->p_matrix[jj1][jj2][completeS].row[testCount[zsrf][subset]] = r;
	      testCount[zsrf][subset]++;
	      systemInfo->p_matrix[jj1][jj2][completeS].norm +=(double)val;
	    }
	  }
	for(zsrf = 0;zsrf<zDim;zsrf++)
	  for(r=0;r<sinoSize;r++)
	    blurredSpace[zsrf][r]=0.0;
	
	// TEST
	for(zsrf = 0;zsrf<zDim;zsrf++)
	  for(subset=0;subset<numSubSets;subset++)
	    if(testCount[zsrf][subset]!=nRows[zsrf][subset])
	      printf("Sparse Error at zsrf = %d, subset = %d, test = %d, nRows = %d\n", zsrf, subset, testCount[zsrf][subset], nRows[zsrf][subset]);
      }
  }
		  
	   


  /* TEST SYSTEM MATRIX */
#ifdef TESTSYS
  {
    FILE *fpt;
    fpt = fopen("testSysval.yes","w");
    
    //    for (jj1=0;jj1<imageInfo->yDim;jj1++)
    //      for (jj2=0;jj2<imageInfo->xDim;jj2++)
	for(subset = 0;subset<numSubSets;subset++)
	  for(zsrf = 0;zsrf<zDim;zsrf++){
	    completeS = (subset)*zDim+zsrf;
	    for(r=0;r<systemInfo->p_matrix[15][15][completeS].nRows;r++)
	      fprintf(fpt,"%f  ", systemInfo->p_matrix[15][15][completeS].val[r]);
	  }
    fclose(fpt);

  
     fpt = fopen("testSysrow.yes","w");
     
     for(subset = 0;subset<numSubSets;subset++)
       for(zsrf = 0;zsrf<zDim;zsrf++){
	 completeS = (subset)*zDim+zsrf;
	 for(r=0;r<systemInfo->p_matrix[15][15][completeS].nRows;r++)
	   fprintf(fpt,"%d  ", systemInfo->p_matrix[15][15][completeS].row[r]);
       }
     fclose(fpt);
     for(subset = 0;subset<numSubSets;subset++){
       for(zsrf = 0;zsrf<zDim;zsrf++){
	 completeS = (subset)*zDim+zsrf;
	 printf("%d, ", systemInfo->p_matrix[15][15][completeS].nRows);
       }
       printf("\n");
     }
     printf("\n\n+++++++++\n\n");
      for(subset = 0;subset<numSubSets;subset++){
       for(zsrf = 0;zsrf<zDim;zsrf++){
	 completeS = (subset)*zDim+zsrf;
	 printf("%f, ", systemInfo->p_matrix[15][15][completeS].norm);
       }
       printf("\n");
      }

     

		
    exit(-2);
  }    
  #endif

    fflush(stderr);
  Print("Cleaning up System Matrix Generation - freeing memory\n");
      fflush(stdout);
      
  free((void*)one_p_column.val);
  free((void*)one_p_column.row);
  multifree((void**)blurredSpace,2);
  multifree((void**)nRows,2);
	
      fflush(stderr);
      Print("Finished generating System Matrix\n");
      fflush(stdout);

      
} // End of generate_system_matrix()

/************************************************************************************************/
/* Function: Pulls column j from system matrix (or computes on the fly) */
//

void system_matrix_j(int im_row, int im_col, int im_subset,
		     struct sino_info *sinoInfo,
		     struct recon_image_info *imageInfo,
		     struct system_matrix_info *systemInfo)
{ 

#ifdef PRECOMPUTE_P_MATRIX    

  systemInfo->P_j = systemInfo->p_matrix[im_subset][im_row][im_col];
  
#else   // Compute P_*j on the fly - do not store P

    {
    static int first = 1;
    int r;
    if(first){
      first = 0;
      fflush(stderr);
      find_pixel_contributions(sinoInfo, imageInfo,&pixelCont);  // fill pixelCont
    }
       
    get_one_p_column_for_j(im_row,im_col,xPixelSize, yPixelSize,tfov,sinoInfo,imageInfo,&pixelCont,&one_p_column);
  for (r=0;r<one_p_column.nRows;r++)
    {
      systemInfo->P_j.val[r] = (float)one_p_column->val[r];
      systemInfo->P_j.col[r] = one_p_column->col[r];
      systemInfo->P_j.nRows = one_p_column->nRows;
    }     
  }
#endif  // End compute P on the fly
    
} // End system_matrix_j()

/************************************************************************************************/
/* Function: Pulls column j from PSF Blurred system matrix  */
//

void system_matrix_j_PSF21(int im_row, int im_col, int im_subset,
		     struct sino_info *sinoInfo,
		     struct recon_image_info *imageInfo,
		     struct system_matrix_info *systemInfo,int plane)
{

  systemInfo->P_j = systemInfo->p_matrix[im_row][im_col][im_subset*systemInfo->srfDim[1]+plane];
 
} // End system_matrix_j_PSF21()

/************************************************************************************************/
/* Function: Compute a column of P for a given pixel jj1,jj2 */
/**/

void get_one_p_column_for_j(int im_row, int im_col,
			     double xPixelSize,double yPixelSize,double tfov,
			    struct sino_info *sinoInfo,
			    struct recon_image_info *imageInfo,
			    double ***pixelCont,
			    struct p_column_dbl *one_p_column)
{

  int	ind,ind_min,ind_max,pr,prof_ind,pind,i,proj_count,len;
  double Pval,t_min,t_max,ang,x,y,t_pix_ctr,t_prj,Dt;
  double max_delta_t;
  double projWidth;
  static int first=1;
  static double *cs,*si;
  double t_0;
  

  len = (int)LEN;
  
  if (first)
    {
      /* precompute cos and sin for all projection angles*/
      first=0;
      cs=(double*)mget_spc(sinoInfo->nAngles,sizeof(double));
      si=(double*)mget_spc(sinoInfo->nAngles,sizeof(double));
      
      for (pr=0; pr < sinoInfo->nAngles; pr++)
	{
	  //	  ang = (double)-M_PI/2.0 + ((double)pr)*(sinoInfo->angleSpacing);
	  ang = ((double)pr)*(sinoInfo->angleSpacing); 
	  cs[pr]=cos((double) ang);
	  si[pr]=sin((double) ang);
	}
    }
  
  projWidth =  sinoInfo->projWidth;


  /* What if pixels are not square ? */
  // max_delta_t = imageInfo->xPixelSize/sqrt(2.0) + 0.5*projWidth;  // Old way, change max_s in find_pixel_contributions()
  max_delta_t = imageInfo->xPixelSize/sqrt(2.0) + 0.5*projWidth;
  
  y = (-tfov+yPixelSize)/2.0 + im_row*yPixelSize;
  x = (-tfov+xPixelSize)/2.0 + im_col*xPixelSize;
  
  proj_count = 0;
  
  // Is t_0 the start of the 1st projection or the center of the first projection :: I think it should be the start of the first projection - center of 1st proj now
  // t_0 = -0.5*((double)(sinoInfo->nProjs))*(double)sinoInfo->projSpacing-(double)sinoInfo->projSpacing/2.0; // center
  if(sinoInfo->nProjs & 1)
    t_0 = -0.5*(double)((sinoInfo->nProjs-1)*sinoInfo->projSpacing);
  else
    t_0 = -0.5*(double)((sinoInfo->nProjs)*sinoInfo->projSpacing);

  //  t_0 = -0.5*((double)(sinoInfo->nProjs))*(double)sinoInfo->projSpacing+(double)sinoInfo->projSpacing/2.0; // center of projection
  

  for(pr=0; pr < sinoInfo->nAngles; pr++)
    {
     
      pind = pr*sinoInfo->nProjs;
      
      /* pixel center is    y*cos((double) ang) - x*sin((double) ang) */
      /* so go down to tmin=center-projWidth/2  and upto  t_max=center+projWidth/2 */
      /*t_min = y*cos((double) ang) - x*sin((double) ang) - delta;
	t_max = t_min + 2.0*delta;*/
      
      t_pix_ctr = y*cs[pr] - x*si[pr];
      t_min = t_pix_ctr - max_delta_t;
      t_max = t_pix_ctr + max_delta_t;
      
      /* if range below first projection */
      if(t_max < t_0) 
	continue;
      
      /* From physical t_min and t_max, compute the 
	 index range of parallel proj. */
  
      ind_min = (int)ceil((t_min-t_0)/sinoInfo->projSpacing); // Round to - inf
      ind_max = (int)((t_max-t_0)/sinoInfo->projSpacing);
      
      ind_min = (ind_min<0) ? 0:ind_min;
      ind_max = (ind_max>=sinoInfo->nProjs) ? sinoInfo->nProjs-1:ind_max;

      for(i = ind_min; i <=ind_max; i++) 
	{
	  
	  ind = pind+i; /* linear index in 1D array 
			   of all proj. (angle*n_t + t)*/
	  
	  /* get profile_index that is closest to this projection */
	  /* profiles are in discrete t-steps of 2pixels/LEN = 
	     2.0*delta / LEN */
	  t_prj = t_0+i*sinoInfo->projSpacing;
	  Dt = fabs(t_prj-t_pix_ctr);

	  //	  prof_ind = (int)(Dt/max_delta_t * (double)LEN + 0.5); 

	  prof_ind = (int)(Dt/max_delta_t * (double)len);   // used to be +0.5
	  

	  
	  if(prof_ind >= len || prof_ind <0)
	    {
	      if(prof_ind == len)
		prof_ind = len-1;
	      else if (prof_ind == -1)
		prof_ind = 0;
	      else
		{
		  printf("y = %f, x = %f, t_min=%f, t_max = %f\n", y,x,t_min,t_max);
		  printf("im_row = %d, im_col = %d, pr = %d\n", im_row, im_col, pr);
		  printf("ang = %f, cs[pr] = %f, si[pr] = %f\n", pr*sinoInfo->angleSpacing, cs[pr],si[pr]);
		  printf("Dt = %f, max_delta_t = %f, len = %d, t_pix_ctr = %f, t_prj = %f\n", Dt, max_delta_t, len, t_pix_ctr, t_prj);
		  Error("Error in get_one_p_column_for_j()\n  prof_ind=%d is out of range [0,%d], ind_min=%d, ind_max = %d, i = %d\n",prof_ind,len-1, ind_min, ind_max,i);
		 
		}
	    }
	  /* Assign output */
	  if((Pval=(*pixelCont)[pr][prof_ind]) > 0.0) 
	    {	
	      one_p_column->val[proj_count] = Pval;  // AMA: Could have lookup here to shape the projection profile accordingly based on ind and ind_middle (variant PSF along distance)! 			      
	      one_p_column->row[proj_count] = ind;     /* Index for ray    */
	      proj_count++;
	    }
	}
    }

  one_p_column->nRows = proj_count;
  
} // End get_one_p_column_for_j()

/************************************************************************************************/
/* Function: Find how pixels at distances [0 to projWidth] from a projection contribute
   to the projection at all angles */
/**/

void find_pixel_contributions(struct sino_info *sinoInfo,struct recon_image_info *imageInfo, double ***pixelCont)
{

  double projWidth, deltaPix, max_s, delta_s;
  double nMicroPixels_dbl, delta_m, xy0_m, A_m, pix_h;
  int ang_index, t_index,i,j,len,pix_subsample;
  double ang,csa, ssa,t_prj, sum,x_m,y_m,t_m,numSubPixelsInProjWidth;

  len = (int)LEN;
  pix_subsample = (int)PIX_SUBSAMPLE;

  *pixelCont = (double**)multialloc(sizeof(double),2,sinoInfo->nAngles,len);
  //  pixelCont = (double **)get_img(len,sinoInfo->nAngles,sizeof(double));
  projWidth = sinoInfo->projWidth;

  deltaPix = imageInfo->xPixelSize;

  //  max_s = deltaPix/sqrt(2.0) + 0.5*projWidth;  // Also change max_delta_t
  max_s = deltaPix/sqrt(2.0) + 0.5*projWidth;
  delta_s = max_s/((double)(len));

  nMicroPixels_dbl=(double)pix_subsample;

  numSubPixelsInProjWidth = (double)(projWidth * nMicroPixels_dbl/deltaPix);  // Not Useful notes:2/25/04  Enforcing same mass for different projWidth is not appropriate...
  
  //  m denotes the micropixels used to subsample a pixel
  delta_m = deltaPix/nMicroPixels_dbl;
  xy0_m = -(nMicroPixels_dbl-1.0)/2.0 * delta_m;
  A_m = delta_m*delta_m; // area of micropix  / 1000.0 to scale correctly-need correct FOV?
  // Print("\nd_m = %f, xy0_m= %f,   A_m = %f\n", delta_m, xy0_m, A_m);

  pix_h = deltaPix/2.0;
  
  for (ang_index=0;ang_index<sinoInfo->nAngles;ang_index++)
    {
      
      ang = sinoInfo->angleSpacing * (double)ang_index;
      
      while(ang >= (double)M_PI/2.0)
	ang -= (double)M_PI/2.0;
      
      csa=cos(ang);
      ssa=sin(ang);
      
      for (t_index=0;t_index<len;t_index++)
	{
	  t_prj = delta_s * (double)t_index;
	  sum=0.0;

	  for (i=0;i<pix_subsample;i++)
	    {
	      y_m = xy0_m + delta_m*(double)i;
	      for (j=0;j<pix_subsample;j++)
		{
		  x_m = xy0_m + delta_m*(double)j;
		  t_m = y_m*csa-x_m*ssa;
		  sum += proj_shape_rect(t_prj-t_m, projWidth);  
		  // sum += proj_shape_triangle(t_prj-t_m, projWidth); // AMA Change projection shape
		}
	    }
	  sum *= A_m;

	  (*pixelCont)[ang_index][t_index] = sum;
	}
    }

  //#define DUMP_PIXELCONT  
#ifdef DUMP_PIXELCONT
  {
    FILE* fp;
    
    Print("Dumping pixel contribution --> \n");
    fp=fopen("pixelCont.yes","w");
    for (ang_index=0;ang_index<sinoInfo->nAngles;ang_index++)
      {
        for (t_index=0;t_index<len;t_index++)
          fprintf(fp,"%g ",(*pixelCont)[ang_index][t_index]);
        fprintf(fp,"\n");
      }
    fclose(fp);
  }
#endif

 } // End find_pixel_contributions()
      



/************************************************************************************************/
/* Functions: Returns value of projection at distance, t  */
/*  */

// Rectangular shape projections 
double proj_shape_rect(double t, double projWidth)
{
  double val;
  // t must be within [projWidth/2 , projWidth/2)
  
  if(t<0.0){
    t = -t;
    if(t<=projWidth/2.0)  // Force non-overlapping, complete projection space
      val = 1.0/projWidth;
    else
      val = 0.0;
  }
  else if(t<projWidth/2.0)
      val = 1.0/projWidth;
  else
    val = 0.0;

  return(val*2.0);
}

// Triangular shape projections
double proj_shape_triangle(double t, double projWidth) 
{
  double val;

  if(t<0.0)
    t = -t;
  
  if(t<=projWidth/2.0)
    {
    val = 2.0/projWidth;
    val = val - val*val*t;
    }
  else
    val = 0.0;

  return(val);
 }



